﻿#chtEnhancedUI.py
#A monkey patching plugin specifically for the zh_TW locale.  Utilizing the new character processing framework, several enhancements are implemented for a more user friendly experience .
#The default verbosity of character descriptions is limited to one description spoken at a time.  User can enumerate through the list of descriptions available in the dictionary by repeated calls to the getCharacterDescription function by double pressing numpad2 or NVDA+, twice quickly for laptop.
#Triple-pressing numPad8 (or NVDA+i three times quickly now speaks the line with one description added as the character description.  Previously, the entire list of descriptions for each character is spoken, making the function not very usable.
#User can enumerate through the list of descriptions in reverse order by pressing Shift+numPad2 or NVDA+Shift+, for laptop.
#The setting of enhancement functionality can be toggled on and off by pressing NVDA+0.  This affects the speech._speakSpellingGen function, so that character descriptions can be heard when moving both the review cursor and system caret.
#When the "Chinese enhanced UI functionality is toggled on, the speaking of hexadecimal unicode representation for review_currentCharacter is disabled, allowing for the user to cycle through the list of descriptions in the character description dictionary more effectively.
#Optionally for now, when the Chinese enhanced UI is toggled on, if the user wishes to disable phonetic spelling of alphabets when typing and for caret and review cursor movements, the list of phonetic alphabets in the zh_TW characterDescriptions.dic needs to be manually deleted.
#Should the user modify the zh_TW characterDescription file, English phonetic alphabets can still be heard when pressing numpad2.
#Written by Michael Chen <m11chen@gmail.com>

import globalPluginHandler
import ui
import config
from config import conf
from speech import *
from globalCommands import *
import characterProcessing

#Create a section in the config file to store dynamic information related to this plugin .
confSect = conf.get("chtEnhancedUI")
if not confSect:
	confSect = conf["chtEnhancedUI"] = {}
conf["chtEnhancedUI"]["enabled"] = "Off"	#Stores the state of the Chinese enhanced UI functionality.
conf["chtEnhancedUI"]["previousCharacter"] = ""	#Stores the character which had its description spoken last.
conf["chtEnhancedUI"]["direction"] =""	#stores the direction in which the list of descriptions is to be enumerated.
conf["chtEnhancedUI"]["descIndex"] = 0	#Stores the position in the list of descriptions for the character last spoken.
conf["chtEnhancedUI"]["alphanumeric"] = False	#stores whether the character which is having its description retrieved is alphanumeric.


def _customSpeakSpellingGen(text,locale,useCharacterDescriptions):
	#Check if the Chinese enhanced UI functionality is turned on.
	#If yes, all scripts and functions which calls the speech._speakSpellingGen function will have the useCharacterDescription passed as True.
	#This is done so when the caret and review cursor moves, the character description is automatically spoken.
	if conf["chtEnhancedUI"]["enabled"] == "On":
		locale="zh_tw"
		useCharacterDescriptions = True

	synth=getSynth()
	synthConfig=config.conf["speech"][synth.name]
	buf=[(text,locale,useCharacterDescriptions)]
	for text,locale,useCharacterDescriptions in buf:
		textLength=len(text)
		for count,char in enumerate(text): 
			uppercase=char.isupper()
			charDesc=None

			if useCharacterDescriptions:
				#do not speak character descriptions for alphanumeric characters unless the function is called by the default review_currentCharacter method.
				#This is to prevent phonetic spelling of the alphabets when typing, and moving the caret and review cursor.
				#The character descriptions for the alphabets need to be removed in the zh_TW characterDescriptions.dic to make this possible.
				if conf["chtEnhancedUI"]["alphanumeric"]:
					#zh_TW character description dictionary no longer have alphabets, so use en as the locale.
					charDesc=characterProcessing.getCharacterDescription("en",char.lower())
				else:
					charDesc=characterProcessing.getCharacterDescription(locale,char.lower())
			if charDesc and conf["chtEnhancedUI"]["enabled"] == "On":
				char=charDesc
			elif charDesc:
				char=charDesc[0] if textLength>1 else u"\u3001".join(charDesc)
			else:
				char=characterProcessing.processSpeechSymbol(locale,char)
			if uppercase and synthConfig["sayCapForCapitals"]:
				char=_("cap %s")%char
			if uppercase and synth.isSupported("pitch") and synthConfig["capPitchChange"]:
				oldPitch=synthConfig["pitch"]
				synth.pitch=max(0,min(oldPitch+synthConfig["capPitchChange"],100))
			index=count+1
			log.io("Speaking character %r"%char)
			speechSequence=[LangChangeCommand(locale)] if config.conf['speech']['autoLanguageSwitching'] else []
			if len(char) == 1 and synthConfig["useSpellingFunctionality"]:
				speechSequence.append(CharacterModeCommand(True))
			if index is not None:
				speechSequence.append(IndexCommand(index))
			speechSequence.append(char)
			synth.speak(speechSequence)
			if uppercase and synth.isSupported("pitch") and synthConfig["capPitchChange"]:
				synth.pitch=oldPitch
			while textLength>1 and (isPaused or getLastSpeechIndex()!=index):
				for x in xrange(2):
					args=yield
					if args: buf.append(args)
			if uppercase and  synthConfig["beepForCapitals"]:
				tones.beep(2000,50)
		args=yield
		if args: buf.append(args)
def modified_getCharacterDescription(locale,character):
	"""
	This function overrides the default getCharacterDescription function.  If the locale is set to a language other than zh_TW, the function returns the same as the default function.
	@param locale: the locale (language[_COUNTRY]) the description should be for.
	@type locale: string
	@param character: the character  who's description should be retrieved.
	@type character: string
	@return:  the found description for the given character.  If the locale is set to zh_TW, it returns a single description instead of the entire list of descriptions found in the locale dictionary.
	@rtype: string
	"""
	try:
		l=characterProcessing._charDescLocaleDataMap.fetchLocaleData(locale)
	except LookupError:
		if not locale.startswith('en'):
			return characterProcessing.getCharacterDescription('en',character)
		raise LookupError("en")
	desc=l.getCharacterDescription(character)
	if (not desc and not locale.startswith('en')) and conf["chtEnhancedUI"]["enabled"]=="Off":
		desc=characterProcessing.getCharacterDescription('en',character)

	if conf["chtEnhancedUI"]["enabled"] == "Off":
		return desc
	else:
		#Update parameters for the previous getCharacterDescription call from the config section .
		char= conf["chtEnhancedUI"]["previousCharacter"]
		direction = conf["chtEnhancedUI"]["direction"]
		index= conf["chtEnhancedUI"]["descIndex"]
		if not desc:
			#There is no description for the character, so return nothing to allow the character to be passed to the processSymbol function.
			#Allows for speaking of punctuation and symbols absent from the dictionary.
			return
#		desc=" ".join(desc)
		#logic for determining the list position.
		if char == character and direction == "Forward":
			index= index+1
			if index == len(desc):
				index=0
		elif char  == character and direction == "Reverse":
			index = index-1
			if index < 0:
				index= len(desc)-1
		else:
			#the caret or review cursor has moved, so reset everything to default.
			#In Chinese, it is customary to use the same character twice in succession(疊字), especially adjectives and adverbs, for emphasis purpose.  Resetting to defaults allows for a more consistent listening experience when moving caret and review cursor quickly through a sentence.
			char  =""
			index= 0
		#List position is determined, so now retrieve the corresponding description.
		currentDesc = desc[index]
		#Store the list index in case of repeated calls on the same character.
		conf["chtEnhancedUI"]["descIndex"] = index
		#Check for possible situations to return:
		#For repeated calls on the same character, append the character behind the description.  This could be desirable when the list of descriptions is long or ambiguous as to which character the description actually pertains to.
		if char == character and len(desc)>1:
			return currentDesc+u'的'+char
		else:
			#The description is for a character different from the one in the previous call, or there is only one description.
			#First store the character to the config file, in case for repeated calls for its other descriptions.
			conf["chtEnhancedUI"]["previousCharacter"]=character
			#First speak the character followed by space to allow for a pause in the synthesizer before the first description is spoken.  This could be desirable when the user moves the caret or review cursor quickly through a sentence and wants to hear the character without any extra information.
		#In the case of having only one description, for Chinese, this usually means that the character is very rare or needs to be described pictorially, so the description itself will suffice.
			return character+" "+currentDesc

class GlobalPlugin(globalPluginHandler.GlobalPlugin):

	def __init__(self):
		super(GlobalPlugin, self).__init__()
		self._oldSpeakSpellingGen = speech._speakSpellingGen
		self.old_getCharacterDescription = characterProcessing.getCharacterDescription
		if conf["chtEnhancedUI"]["enabled"] == "On":
			speech._speakSpellingGen = _customSpeakSpellingGen
			characterProcessing.getCharacterDescription = modified_getCharacterDescription

	def script_ToggleChtEnhancedUI(self,gesture):
		if conf["chtEnhancedUI"]["enabled"] == "Off":
			conf["chtEnhancedUI"]["enabled"] = "On"
		else:
			conf["chtEnhancedUI"]["enabled"] = "Off"
		ui.message(_("Chinese enhanced UI "+conf["chtEnhancedUI"]["enabled"]))
	script_ToggleChtEnhancedUI.__doc__=_("Toggles on or off the Chinese enhanced UI functionality.")

	def script_forward_review_currentCharacter(self,gesture):
		conf["chtEnhancedUI"]["direction"]="Forward"
		info=api.getReviewPosition().copy()
		info.expand(textInfos.UNIT_CHARACTER)
		#Check the state of the Chinese enhanced UI functionality.
		#If turned on, hardcode the scriptCount to 1 to bypass speaking the character by itself and disable speaking hexadecimal representation of the character.
		if conf["chtEnhancedUI"]["enabled"]=="On":
			scriptCount = 1
		else:
			#Perform default action of review_currentCharacter script, so start counting repeated script calls.
			scriptCount=scriptHandler.getLastScriptRepeatCount()
		try:
			c = ord(info.text)
		except:
			c=0
		if scriptCount==0:
			speech.speakTextInfo(info,unit=textInfos.UNIT_CHARACTER,reason=speech.REASON_CARET)
		elif scriptCount==1:
			#Check if the character is alphanumeric
			if c>=65 and c<=122:
				conf["chtEnhancedUI"]["alphanumeric"] = True
			speech.speakSpelling(info.text,useCharacterDescriptions=True)
		else:
			try:
				speech.speakMessage("%d," % c)
				speech.speakSpelling(hex(c))
			except:
				speech.speakTextInfo(info,unit=textInfos.UNIT_CHARACTER,reason=speech.REASON_CARET)
		#Reset parameters to prepare for the next call to retrieve other descriptions.
		conf["chtEnhancedUI"]["alphanumeric"] = False
		conf["chtEnhancedUI"]["direction"] = ""
	script_forward_review_currentCharacter.__doc__=_("Reports the character of the current navigator object where the review cursor is situated. If this key is pressed twice, character descriptions are spoken.  If pressed three times, ascii and hexadecimal representations for the character are spoken")

	def script_reverse_review_currentCharacter(self,gesture):
		conf["chtEnhancedUI"]["direction"] = "Reverse"
		info=api.getReviewPosition().copy()
		info.expand(textInfos.UNIT_CHARACTER)
		try:
			c = ord(info.text)
		except:
			c=0
		if c>=65 and c<=122:
			conf["chtEnhancedUI"]["alphanumeric"] = True
		speech.speakSpelling(info.text,useCharacterDescriptions=True)
		conf["chtEnhancedUI"]["alphanumeric"] = False
		conf["chtEnhancedUI"]["direction"] = ""
	script_reverse_review_currentCharacter.__doc__=_("If the locale is set to zh_TW, enumerates in reverse order through the list of nouns in the Chinese character descriptions dictionary.")

	def terminate(self):
		speech._speakSpellingGen = self._oldSpeakSpellingGen
		characterProcessing.getCharacterDescription = self.old_getCharacterDescription
		super(GlobalPlugin, self).terminate()

#: Dictionary for key bindings
	__gestures = {
		"kb:NVDA+0": "ToggleChtEnhancedUI",
		"kb:numPad2": "forward_review_currentCharacter",
		"kb:Shift+numPad2": "reverse_review_currentCharacter",
		"kb(laptop):NVDA+,": "forward_review_currentCharacter",
		"kb(laptop):NVDA+Shift+,": "reverse_review_currentCharacter",
	}
